/**
 * Module dependencies.
 */

require('dotenv').config();

const express = require('express');
const fs = require("fs");

const http = require('http');
const path = require('path');
const ngrok = require("ngrok");
const request = require("request");

const AWS = require('aws-sdk');

AWS.config.update({
    region : process.env.AWS_REGION,
    endpoint: process.env.AWS_DYNAMO_DB_ENDPOINT
});

const YAML = require('yamljs');

var app = express();

var handler = require(__dirname + "/index")(process.env.HOME_DIR || __dirname);

// all environments
app.set('port', process.env.PORT || 4000);
app.use(express.favicon());
app.use(express.logger('dev'));

app.use(function(req, res, next) {
  req.rawBody = '';
  req.setEncoding('utf8');

  req.on('data', function(chunk) { 
    req.rawBody += chunk;
  });

  req.on('end', function() {
    next();
  });
});

app.use(express.methodOverride());
app.use(app.router);

const staticpath = path.join(process.env.HOME_DIR, process.env.BOT_UNIQUE_ID, "assets")
app.use(`/${process.env.BOT_UNIQUE_ID}/v1/assets`, express.static(staticpath));

// development only
if ('development' == app.get('env')) {
  app.use(express.errorHandler());
}

var handleRequest = (req, res)=>{
  var context = {
    functionName : req.params["uid"],
    succeed : function(result) {
        if (result.statusCode === 200){
          if (result.headers["Content-Type"] === "application/json"){
            console.log(`DEBUG: ${JSON.stringify(result.body)}`);
            
            return res.json(result.body);   
          }
          else if (result.headers["Content-Type"] === "text/plain"
          || result.headers["Content-Type"] === "application/xml"){
            console.log(`DEBUG: ${result.body}`);

            return res.send(result.body);   
          }
          return res.send(result.body);
        }
        return res.send(result.statusCode);
    },
    fail : function(err){
        if (err){
          console.log("DEBUG: " + JSON.stringify(err));
          res.send(err.code || 500);
        }
    }
  };
  // AWS Lamda Compatible function.
  handler.bind({
    httpMethod : req.method,
    body : req.rawBody,
    queryStringParameters : req.query,
    path : req.path,
    host : req.host

  }, context, function(err, result){
      if (err !== null){
        res.send(err);
      }
      else{
        res.send(result);
      }
  });
};

function fixHeader(req, res, next) {
  res.removeHeader('X-Powered-By');
  res.header("ETag", process.env.BOT_UNIQUE_ID)
  res.header('Accept-Ranges', 'bytes');
  next();
}

app.get('/:uid/v1/assets/:name', fixHeader, (req, res)=>{
    // facebook &
    const name = req.params.name.replace(/\&[^$]+/ig, "")
    const root = path.join(process.env.HOME_DIR, process.env.BOT_UNIQUE_ID, "assets");
    res.sendfile(name, {
      root: root,
      cacheControl : false
    });
});
app.get("/:uid/v1", handleRequest);
app.post("/:uid/v1", handleRequest);

let dynamoDB = new AWS.DynamoDB();

var params = {
        TableName : `bot-${process.env.AWS_LAMBDA_FUNCTION_NAME}-key_values`,
        KeySchema: [
            { AttributeName: "key", KeyType: "HASH"},
        ],
        AttributeDefinitions: [
            { AttributeName: "key", AttributeType: "S" },
        ],
          ProvisionedThroughput: {
            ReadCapacityUnits: 5,
            WriteCapacityUnits: 5
        }
};

dynamoDB.createTable(params, (err, data)=>{
    if (!err){
        console.log(data);
    }
});

function preprocessParams(params){
    Object.keys(params).forEach(function(name) {
        var param = params[name]
        if (param &&
            typeof param === 'object') {
            params[name] = JSON.stringify(param)
        }
    });
    return params;
}

function setupFacebookWebhook(callbackUrl){
   return new Promise((resolve, reject)=>{
      request({
          "url" : `https://graph.facebook.com/${process.env.RECIME_FACEBOOK_APP_ID}/subscriptions`,
          "method": "POST",
          qs : {
            access_token : `${process.env.RECIME_FACEBOOK_APP_ID}|${process.env.RECIME_FACEBOOK_APP_SECRET}`
          },
          json : true,
          form : preprocessParams({
            object : "page",
            "callback_url" :callbackUrl, 
            fields : "messages,messaging_postbacks,messaging_optins,message_deliveries,message_reads",
            verify_token: process.env.BOT_UNIQUE_ID
          })
      }, (err, response, body)=>{
          if(err){
            reject(err);
            return;
          }
          if (body.error){
            reject(body.error);
            return;
          }
          resolve();
      });
   });
}

function nlpConfigure(){
    return new Promise((resolve, reject)=>{
      request({
          method : "POST",
          json : true,
          url : `https://graph.facebook.com/v2.8/me/nlp_configs?nlp_enabled=true&&custom_token=${process.env.WIT_TOKEN}`,
          qs : {
            access_token : process.env.RECIME_FACEBOOK_ACCESS_TOKEN
          }
      }, (err, response, body)=>{
        if (err){
            reject(err);
            return;
        }
        resolve(`[FACEBOOK-NLP] ${JSON.stringify(body)}`);
      });
    });
}

function whitelistFBDomain(domains){
    return new Promise((resolve, reject)=>{
      request({
          method : "POST",
          json : true,
          url : "https://graph.facebook.com/v2.6/me/messenger_profile?",
          qs : {
            access_token : process.env.RECIME_FACEBOOK_ACCESS_TOKEN
          },
          body : {
            whitelisted_domains : domains
          }
      }, (err, response, body)=>{
        if (err){
            reject(err);
            return;
        }
        resolve(`[Domain whitelist] ${JSON.stringify(body)}`);
      });
    });
}

function subscribeToPage(){
    return new Promise((resolve, reject)=>{
      request({
          method : "POST",
          json : true,
          url : "https://graph.facebook.com/v2.6/me/subscribed_apps",
          qs : {
            access_token : process.env.RECIME_FACEBOOK_ACCESS_TOKEN
          }
      }, (err, response, body)=>{
        if (err){
            reject(err);
            return;
        }
        resolve(`[FACEBOOK-PAGE] ${JSON.stringify(body)}`);
      });
    });
}

//curl -F url=PASTE_YOUR_BOT_URL https://api.telegram.org/botPASTE_YOUR_ACCESS_TOKEN_HERE/setWebhook
function setTelegramWebhook(url){
  return new Promise((resolve, reject)=>{
      request({
        json : true,
        url :`https://api.telegram.org/bot${process.env.RECIME_TELEGRAM_ACCESS_TOKEN}/setWebhook`,
        form: preprocessParams({
          url : url
        })
      }, (err, response, body)=>{
          if (err){
            reject(err);
            return;
          }
          resolve(`[TELEGRAM] ${JSON.stringify(body)}`);
      });
  });
}

function startDebugger(options){
  ngrok.connect(options, function(err, url){
      if(err){
          console.error(err);
      }
  
      const endpoint = `${url}/${process.env.BOT_UNIQUE_ID}/v1`;

      console.log("####");
    
      console.log(endpoint);

      if (process.env.RECIME_VIBER_ACCESS_TOKEN){
          request({
            json : true,
            method : "POST",
            url : " https://chatapi.viber.com/pa/set_webhook",
            headers : {
              "X-Viber-Auth-Token": process.env.RECIME_VIBER_ACCESS_TOKEN
            },
            body : {
              url : endpoint,
              event_types: ["delivered", "seen", "failed", "subscribed", "unsubscribed", "conversation_started"]
            }
          }, (err, response, body)=>{
              if (err){
                console.log(err);
              }
          });
      }
      // configure facebook webook.
      if (process.env.RECIME_FACEBOOK_APP_ID && process.env.RECIME_FACEBOOK_APP_SECRET)
          setupFacebookWebhook(endpoint)
            .then(subscribeToPage)
            .then(console.log)
          .catch(console.error);
      if (process.env.RECIME_FACEBOOK_ACCESS_TOKEN && process.env.WIT_TOKEN)
        nlpConfigure()
          .then(console.log)
        .catch(console.error)
      // if (process.env.RECIME_FACEBOOK_ACCESS_TOKEN)
      //   whitelistFBDomain([
      //       require("url").parse(url).host
      //   ]).then(console.log)
      //   .catch(console.error)
      if (process.env.RECIME_TELEGRAM_ACCESS_TOKEN)
        setTelegramWebhook(endpoint)
          .then(console.log)
        .catch(console.error)
  });
}

http.createServer(app).listen(app.get('port'), function(){
  if (process.env.BOT_UNIQUE_ID){
      console.log("[INFO] Debugger is started at:")
      console.log('http://localhost:' + app.get('port') + "/" + process.env.BOT_UNIQUE_ID + "/v1");
      console.log("");

      const configPath = path.join(process.env.HOME_DIR, process.env.BOT_UNIQUE_ID, "ngrok.yml");
      
      if (fs.existsSync(configPath)){
        const config = fs.readFileSync(configPath, "utf-8");
        const obj = YAML.parse(config);

        startDebugger({ 
          addr : app.get('port'), 
          authtoken : obj.authtoken,
          subdomain: obj.subdomain
        });
        return;
      } 
      startDebugger({ addr : app.get('port') });
  }else {
      console.log('http://localhost:' + app.get('port'));
  }
});