'use strict';

const fs = require("fs");
const query = require("querystring");
const data2xml = require('data2xml');
const HandleBars = require("handlebars");

const Dialog = require("./dist/dialog").Dialog;
const Script = require("./dist/engine/script").Script;
const Facebook = require("./dist/facebook").Facebook;
const Slack = require("./dist/slack").Slack;
const Telegram = require("./dist/telegram").Telegram;
const Twilio = require("./dist/twilio").Twilio;
const WeChat = require("./dist/wechat");
const Viber = require("./dist/viber").Viber;
const XML = require('./dist/xml').XML;
const SchemaValidator = require('./dist/schema-validator').SchemaValidator;
const MsgObject = require("./dist/msg-object").MsgObject;
const MessageConverter = require("./dist/message-converter").MessageConverter;
const MainModule = require("./dist/engine/main-module").MainModule;
const Logger = require("./dist/logger").Logger;
    
module.exports = (dir)=> {    
    var env = process.env;
    var host = "";
   
    const handler = function(uid, data, platform, success, error){
        const script = new Script();
        const converter = new MessageConverter(platform);

        var options = {
            homedir : dir,
            uid : uid,
            data : converter.convert(data),
            env: env,
            platform : platform,
            host : host
        };
        
        script.execute(options).then(success, error);
    };


    const res = (code, body, contentType)=>{
        return {
            "statusCode" : code,
            "headers": {
                "Content-Type" : contentType || "application/json",
                "Access-Control-Allow-Origin": "*"
            },
            "body": body || (contentType === "application/json"? JSON.stringify({ 
                success : true 
            }) : "")
        }
    };

    const html = (body)=>{
        return res(200, body, "text/html");
    };

    const bind = (event, context, callback) => {
        const body = MsgObject.parse(event.body || '');
        const query = event.queryStringParameters || {};

        host = event.host;

        const cb = (uid, body)=>{
            handler(uid, body, "web", (result)=>{
                return context.succeed(res(200, JSON.stringify(result)));
            }, (err)=>{
                return context.succeed(res(200, JSON.stringify(err)))
            });
        };

        var params = {
            uid : context.functionName,
            homedir : dir,
        }

        if (event.httpMethod === 'GET') {
            if (query['hub.mode'] && query['hub.mode'].match(/subscribe/ig)){
                const facebook = new Facebook(params);

                facebook.verifyToken(query).then((result)=>{
                    context.succeed(res(200, result, "text/plain"));
                }, (err)=> {
                    context.succeed(res(403));
                });
            }
            else if (query['signature'] && query['timestamp'] && query['nonce']){
                const wechat = new WeChat(params);
                if (wechat.verify(query)){
                    context.succeed(res(200, query['echostr'], "text/plain"));
                } else {
                    context.succeed(res(403));
                }
            }else {
                if (query["code"] && process.env.RECIME_SLACK_CLIENT_ID){
                    const slack = new Slack(params);
                    const mainModule = new MainModule(context);
                    slack.authorize(query["code"]).then((result)=>{
                        if (result.success){
                            fs.readFile(slack.resolveSuccessDialogPath(mainModule), 'utf-8', (err, code)=>{
                                const packagePath = path.join(mainModule.dir, "package.json");
                                fs.readFile(packagePath, 'utf-8', (err, body)=>{
                                        // file error
                                        if (err){
                                            return context.succeed(res(500, err));
                                        }
                                        const metadata = JSON.parse(body);
                                        const iconPath = path.join(mainModule.dir, "icon.png");
                                        fs.readFile(iconPath, (err, icon)=>{
                                            const template = HandleBars.compile(code);
                                            const data = {
                                                "title": metadata.name,
                                                "id": context.functionName,
                                                "team" :  {
                                                    name : result.data.team_name,
                                                    id : result.data.team_id
                                                },
                                                "icon" : `data:image/jpeg;base64,${new Buffer(icon).toString("base64")}`
                                            };
                                            context.succeed(html(template(data)));
                                        });

                                });
                            });
                        } else {
                            context.succeed(res(200, result));
                        }
                    });
                }
                else if (query["ref"]){
                    return handler(context.functionName, {
                        sender : query["ref"]
                    }, "web", (result)=>{
                        return context.succeed(html(result));
                    }, (err)=>{
                        return context.succeed(res(500, err));
                    });
                } 
                else {
                    Dialog.html({
                        uid: context.functionName,
                        filepath : `${__dirname}/success-dialog.html`,
                        homedir : dir
                    }).then ((result)=>{
                        context.succeed(html(result));
                    }, (err)=>{
                        context.succeed(res(500, err));
                    });
                }
            }
        } else {
            // convert xml to json if necessary.
            XML.ifConvert(body).then((body)=>{
                SchemaValidator.findOne(body).then((platform)=>{
                    switch(platform){
                        case "facebook":{
                            params.platform = platform;
                            Logger.init(params).then((logger)=>{
                                const facebook = new Facebook({
                                    homedir : params.homedir,
                                    logger : logger,
                                    uid : params.uid
                                });
                                facebook.execute(body, query, handler).then((body)=>{
                                    context.succeed(res(200, JSON.stringify(body)));
                                }, (err)=>{
                                    console.error(err);
                                    context.fail(err);
                                });
                            });
                            break;
                        }
                        case "telegram": {
                            const telegram = new Telegram(params);
                            telegram.execute(body, query, handler).then((body)=>{
                                context.succeed(res(200));
                            }, (err)=>{
                                console.error(err);
                                context.fail(err);
                            });
                            break;
                        }
                        case "slack":{
                            const slack = new Slack(params);
                            slack.verifyToken(body).then((result)=>{
                                context.succeed(res(200, body.challenge, "text/plain"));
                            }, (err)=>{
                                context.succeed(res(403));
                            });
                            break;
                        }
                        case "slack-command":{
                            context.succeed(res(200));
                            break;
                        }
                        case "slack-event":{
                            const slack = new Slack(params);
                            slack.execute(body, query, handler).then ((body)=>{
                                context.succeed(res(200, JSON.stringify(body)));
                            }, (err)=>{
                                console.error(err);
                                context.fail(err);
                            });
                            break;
                        }
                        case "twilio": {
                            const twilio = new Twilio(params);

                            twilio.execute(body, query, handler).then((body)=>{
                                const convert = data2xml();
                                context.succeed(res(200, convert("Response", body), "application/xml"));
                            }, (err)=>{
                                console.error(err);
                                context.fail(err);
                            });
                            break;
                        }
                        case "wechat": {
                            const wechat = new WeChat(params);
                            wechat.execute(body, query, handler).then((body)=>{
                                if (body.MsgId){
                                    const convert = data2xml();
                                    context.succeed(res(200, convert("xml", body), "application/xml"));
                                } else {
                                    context.succeed(res(200, "", "text/plain"));
                                }
                            }, (err)=>{
                                console.error(err);
                                context.fail(err);
                            });
                            break;
                        }
                        case "viber":{
                            if (body.event === "message" || body.event === "conversation_started") {
                                const viber = new Viber(params);
                                viber.execute(body, query, handler).then ((body)=>{
                                    context.succeed(res(200, JSON.stringify(body)));
                                }, (err)=>{
                                    context.succeed(res(200, err.message));
                                });
                            } else {
                                context.succeed(res(200,JSON.stringify({
                                    success :true
                                })));
                            }
                            break;
                        }
                        default : {
                            return cb(context.functionName, body);
                        }
                    }
                }).catch((err)=>{
                    console.error(err);
                    context.fail(err);
                });
            });
        }
    };
    
    return {
        bind : bind
    };
};
